from collections import defaultdict, namedtuple
from io import StringIO
import logging

# TODO(amin): Convert symbol tuple to element name or atomic number tuple

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())

ELEMENTS = {
    'H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg', 'Al',
    'Si', 'P', 'S', 'Cl', 'Ar', 'K', 'Ca', 'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe',
    'Co', 'Ni', 'Cu', 'Zn', 'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr', 'Y',
    'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 'In', 'Sn', 'Sb',
    'Te', 'I', 'Xe', 'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd',
    'Tb', 'Dy', 'Ho', 'Er', 'Tm', 'Yb', 'Lu', 'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir',
    'Pt', 'Au', 'Hg', 'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac',
    'Th', 'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm', 'Md', 'No',
    'Lr', 'Rf', 'Db', 'Sg', 'Bh', 'Hs', 'Mt', 'Ds', 'Rg', 'Cn', 'Nh', 'Fl',
    'Mc', 'Lv', 'Ts', 'Og'
}


# TODO(amin): Use optional caching/memoization to improve performance
# TODO(amin): Support appostrophies
# TODO(amin): Add option to require no repeated symbols
def spell(word, symbols=ELEMENTS):
    """Return a list of any possible ways to spell a word with a given
    set of symbols.

    Example:
    >>> spell('amputation')
    [('Am', 'Pu', 'Ta', 'Ti', 'O', 'N'), ('Am', 'P', 'U', 'Ta', 'Ti', 'O', 'N')]
    """
    log.info('Word: {}'.format(word))

    g = Graph()
    build_spelling_graph(word, g)

    elemental_spellings = sorted(
        [
            tuple(node.value.capitalize() for node in path)
            # There will only ever be at most 2 firsts and 2 lasts.
            for first in g.firsts()
            for last in g.lasts()
            for path in find_all_paths(g._children_of, first, last)
        ],
        reverse=True
    )

    log.info('Spellings: {}'.format(elemental_spellings))

    return elemental_spellings


class Graph():
    """A directed acyclic graph that stores all possible elemental
    spellings of a word.
    """
    def __init__(self):
        self._parents_of = defaultdict(set)
        self._children_of = defaultdict(set)

    def firsts(self):
        """Return nodes with no parents."""
        return self._children_of[None]

    def lasts(self):
        """Return nodes with no children."""
        return self._parents_of[None]

    def add_edge(self, parent, child):
        """Add a parent-child relatonship to the graph. None is ok as a
        key, but not a value.
        """
        log.debug('add_edge({}, {})'.format(parent, child))
        if parent is not None:
            self._parents_of[child].add(parent)
        if child is not None:
            self._children_of[parent].add(child)

    def nodes(self, connected_only=True):
        """Return a list of all nodes."""
        if connected_only:
            return set(
                node for node in
                set(self._children_of.keys()) | set(self._parents_of.keys())
                if node is not None
            )
        else:
            return set(
                node for node in
                set(self._children_of.keys())
                | set(self._parents_of.keys())
                | set(v for s in self._children_of.values() for v in s)
                | set(v for s in self._parents_of.values() for v in s)
                if node is not None
            )

    def edges(self):
        """Return a list of all parent-child relationships."""
        return [
            (parent, child)
            for parent in self._children_of
            for child in self._children_of[parent]
        ]

    # TODO(amin): Eliminate dead-end nodes from exported graph.
    def export(self, connected_only=True):
        """Print a string to stdout that can be piped to Graphviz to
        generate a graph diagram.
        """
        export = StringIO()
        export.write('digraph G {\n')
        export.write('    graph [rankdir=LR];\n')
        export.write('    node [width=0.75 shape=circle];\n')

        edges = [
            (p, c)
            for p, c in self.edges()
            if p is not None and c is not None
        ]
        for parent, child in sorted(edges):
            export.write('    "{}" -> "{}";\n'.format(parent, child))

        for node in sorted(self.nodes(connected_only=connected_only)):
            export.write('    "{}" [label="{}"];\n'.format(node, node.value.capitalize()))
        export.write('}')
        return export.getvalue()


# A single node of the graph. A glyph and its position in the word.
Node = namedtuple('Node', ['value', 'position'])


def build_spelling_graph(word, graph, symbols=ELEMENTS):
    """Given a word and a graph, find all single and double-character
    glyphs in the word. Add them to the graph only if they are present
    within the given set of allowed symbols.
    """

    def pop_root(remaining, position=0, previous_root=None):
        """Pop the single and double-character roots off the front of a
        given string, then recurse into what remains.

        For the word 'because', the roots and remainders for each call
        look something like:

            'b' 'ecause'
                'e' 'cause'
                    'c' 'ause'
                        'a' 'use'
                            'u' 'se'
                                's' 'e'
                                    'e' ''
                                'se' ''
                            'us' 'e'
                        'au' 'se'
                    'ca' 'use'
                'ec' 'ause'
            'be' 'cause'

        For each root present in the set of allowed symbols, add an edge
        to the graph:

            previous root --> current root.

        Keep track of processed values for `remaining` and do not
        evaluate a given value more than once.

        Keep track of the position of each root so that repeated roots
        are distinct nodes.
        """
        if remaining == '':
            graph.add_edge(previous_root, None)
            return

        single_root = Node(remaining[0], position)
        if single_root.value.capitalize() in symbols:
            graph.add_edge(previous_root, single_root)

            if remaining not in processed:
                pop_root(remaining[1:], position + 1, previous_root=single_root)

        if len(remaining) >= 2:
            double_root = Node(remaining[0:2], position)
            if double_root.value.capitalize() in symbols:
                graph.add_edge(previous_root, double_root)

                if remaining not in processed:
                    pop_root(remaining[2:], position + 2, previous_root=double_root)
        processed.add(remaining)

    processed = set()
    pop_root(word)


def find_all_paths(parents_to_children, start, end, path=[]):
    """Return a list of all paths through a graph from start to end.
    `parents_to_children` is a dict with parent nodes as keys and sets
    of child nodes as values.

    Based on https://www.python.org/doc/essays/graphs/
    """
    path = path + [start]
    if start == end:
        return [path]
    if start not in parents_to_children.keys():
        return []
    paths = []
    for node in parents_to_children[start]:
        if node not in path:
            newpaths = find_all_paths(parents_to_children, node, end, path)
            for newpath in newpaths:
                paths.append(tuple(newpath))
    return paths


if __name__ == '__main__':
    test_word = 'Mockery'
    print('{}:\n{}'.format(test_word, spell(test_word)))



import io
import os
import re
from setuptools import setup


here = os.path.abspath(os.path.dirname(__file__))


def read(*names, **kwargs):
    with io.open(
        os.path.join(here, *names),
        encoding=kwargs.get('encoding', 'utf8')
    ) as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]",
        version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string.')


def readme():
    with io.open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
        return f.read()


setup(
    name='Stoichiograph',
    version=find_version('stoichiograph', '__init__.py'),
    description=(
        'Spell words with elemental symbols from the periodic table ("He", "Cu", etc).'
    ),
    long_description=readme(),
    url='https://github.com/mesbahamin/stoichiograph',
    author='Amin Mesbah',
    author_email='mesbahamin@gmail.com',
    license='MIT',
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Education',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Text Processing :: Filters',
        'Topic :: Utilities',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: Microsoft :: Windows :: Windows 7',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    keywords='command-line chemistry words fun combinatorics spelling',
    packages=['stoichiograph'],
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'stoichiograph = stoichiograph.stoichiograph:main',
        ],
    },
    install_requires=[
        'appdirs>=1.4.0',
        'openpyxl>=2.3.5',
        # NOTE(amin): Using syntax introduced in PEP508,
        # install PyQt5 only if using Python 3.5 or greater.
        # 'PyQt5>=5;python_version>="3.5"' should be possible,
        # but currently isn't, so PyQt5 is instead specified
        # here as a reference to an optional dependency:
        'chronophore[qt] ; python_version>="3.5"',
        'SQLAlchemy>=1.1.2'
    ],
)



#!/usr/bin/python3

import argparse
import collections
import json
import logging
import pathlib

from stoichiograph import speller

__title__ = 'stoichiograph'
__author__ = 'Amin Mesbah'
__version__ = '0.0.1'
__description__ = 'Spell words with elemental symbols from the periodic table.'


def get_args():
    parser = argparse.ArgumentParser(
        prog=__title__,
        description=__description__
    )
    parser.add_argument(
        'words',
        help='word(s) for which to find elemental spellings',
        type=str,
        nargs='*'
    )
    parser.add_argument(
        '-b', '--batch-file',
        help='text file containing one word per line'
    )
    parser.add_argument(
        '-c', '--clobber', action='store_true',
        help='overwrite output file if it exists'
    )
    parser.add_argument(
        '--debug', action='store_true',
        help='print debug log'
    )
    parser.add_argument(
        '--list-elements', action='store_true',
        help='print list of elemental symbols and exit'
    )
    parser.add_argument(
        '--export-graph', action='store_true',
        help='export graph of first word as dot code'
    )
    parser.add_argument(
        '-o', '--output-file',
        help='path of output json file'
    )
    parser.add_argument(
        '-s', '--sort', action='store_true',
        help='sort words by length'
    )
    parser.add_argument(
        '-t', '--tuples', action='store_true',
        help='display spellings as tuples'
    )
    parser.add_argument(
        '-v', '--verbose', action='store_true',
        help='print a detailed log'
    )
    parser.add_argument(
        '-V', '--version', action='store_true',
        help='print version info and exit'
    )

    return parser.parse_args()


def main():
    args = get_args()

    if args.version:
        print('{} {}'.format(__title__, __version__))
        raise SystemExit

    if args.list_elements:
        print('{} Elements:'.format(len(speller.ELEMENTS)))
        print(sorted(list(speller.ELEMENTS)))
        raise SystemExit

    if args.debug:
        CONSOLE_LOG_LEVEL = logging.DEBUG
    elif args.verbose:
        CONSOLE_LOG_LEVEL = logging.INFO
    else:
        CONSOLE_LOG_LEVEL = logging.WARNING

    logging.basicConfig(level=CONSOLE_LOG_LEVEL)
    logging.debug('{} {}'.format(__title__, __version__))

    SORT_WORDS = args.sort
    TUPLES = args.tuples

    if args.output_file:
        OUTPUT_FILE = pathlib.Path(args.output_file)
        CLOBBER = args.clobber

        if not CLOBBER and OUTPUT_FILE.exists():
            logging.warning(
                "{} exists. To overwrite, use '--clobber'.".format(OUTPUT_FILE)
            )
            raise SystemExit

    if args.batch_file:
        words_file = pathlib.Path(args.batch_file)
        with words_file.open('r') as f:
            dictionary = f.readlines()

        # TODO(amin): Handle punctuation, apostraphies, etc.
        words = [word.rstrip('\n') for word in dictionary if "'" not in word]
    else:
        words = args.words

    if SORT_WORDS:
        words.sort(key=len, reverse=True)

    if args.export_graph and words:
        g = speller.Graph()
        speller.build_spelling_graph(words[0], g)
        print(g.export())
        raise SystemExit

    spellable = collections.OrderedDict()

    for word in words:
        if TUPLES:
            spellings = speller.spell(word)
        else:
            spellings = [''.join(s) for s in speller.spell(word)]

        if spellings:
            spellable[word] = spellings
            for spelling in spellings:
                print(spelling)

    if args.output_file:
        with OUTPUT_FILE.open('w') as f:
            json.dump(spellable, f, indent=4, sort_keys=False)

    logging.debug('Done!')


if __name__ == '__main__':
    main()



__title__ = 'stoichiograph'
__version__ = '0.1.0'
__license__ = 'MIT'
__author__ = 'Amin Mesbah'
__email__ = 'mesbahamin@gmail.com'
__description__ = 'Spell words with elemental symbols from the periodic table ("He", "Cu", etc).'
